import elf;
import elfstrtab;
import std.stdio;
import std.stream;
import std.system;
import std.algorithm;
import std.iterator;

//"definition symbol types": symbols that are program entities defined
//by source code (as opposed to section or file symbols).
string[int] DEF_SYM_TYPES()
{
    return [
        STT_OBJECT : cast(string)"object",
        STT_FUNC : "function",
        STT_COMMON : "common",
        STT_TLS : "tls"
    ];
}

template StructArrSect(T)
{
    T[] readFromStream(Stream stream, int num, ulong offset, int entitySize)
    {
	if(T.sizeof != entitySize)
	{
	    throw new Error(T.stringof ~ " size differs from the entity size recorded in the ELF file");
	}
	
	stream.position(offset);
	T[] ret = new T[num];
	stream.readExact(&ret[0], ret.length * T.sizeof);
	
	return ret;
    }
}

class ElfSymTab
{
    ElfStrTab secstrtab; //section name string table
    ElfStrTab symstrtab; //symbol name string table
    Elf32_Shdr[] secs; //section headers
    Elf32_Shdr[] ssecs; //target-allocated section headers sorted by base address
    Elf32_Sym[] syms; //symbol definitions
    Elf32_Sym[] ssyms; //program symbol definitions sorted by base address
    
    private void failif(bool cond, lazy string cause)
    {
	if(cond)
	{
	    throw new Error(cause);
	}
    }
    private void checkHeader(Elf32_Ehdr ehdr)
    {
	//ELF?
        failif(ehdr.e_ident[0 .. 4] != ELFMAG, "bad ELF magic number");
	//satisfies our lazy assumptions (32b, host endianness)?
        failif(ehdr.e_ident[EI_CLASS] != ELFCLASS32, "only 32-bit ELF files are supported");
	int data = ehdr.e_ident[EI_DATA];
	failif(data == ELFDATA2MSB && std.system.endian == Endian.LittleEndian,
	       "big endian files can't be read on a little endian host");
	failif(data == ELFDATA2LSB && std.system.endian == Endian.BigEndian,
	       "little endian files can't be read on a big endian host");
    }
    private Elf32_Shdr shdr(int index, lazy string what)
    {
	failif(index == SHN_UNDEF, what ~ " index undefined");
	failif(index < 0 || index >= secs.length, what ~ " index out of bounds");
	return secs[index];
    }
    static bool isTargetSection(Elf32_Shdr shdr)
    {
	return (shdr.sh_flags & SHF_ALLOC) != 0;
    }
    void parseStream(Stream stream)
    {
        Elf32_Ehdr ehdr;
        stream.readExact(&ehdr, ehdr.sizeof);
	checkHeader(ehdr);

	//read section headers
	secs = StructArrSect!(Elf32_Shdr).readFromStream(stream, ehdr.e_shnum, ehdr.e_shoff, ehdr.e_shentsize);
	//read section strtab
	secstrtab = ElfStrTab.readFromStream(stream, shdr(ehdr.e_shstrndx, "section strtab"));
	//find the symbol table header
	foreach(Elf32_Shdr h ; secs)
	{
	    if(h.sh_type == SHT_SYMTAB)
	    {
		//read symbol table definitions
	        syms = StructArrSect!(Elf32_Sym).readFromStream(stream, h.sh_size/h.sh_entsize, h.sh_offset, h.sh_entsize);
		//read symbol string table (in symtab sections, sh_link points to the strtab section)
		symstrtab = ElfStrTab.readFromStream(stream, shdr(h.sh_link, "symbol strtab"));
	    	break;
	    }
	}
	failif(syms == null, "stripped ELF - no symbol table found");
	auto dst = DEF_SYM_TYPES();
	ssecs = filter!(isTargetSection)(secs);
	ssyms = filter!(delegate(Elf32_Sym a) { 
		   return ELF32_ST_TYPE(a.st_info) in dst && (a.st_shndx == SHN_UNDEF || isTargetSection(section(a)));
		})(syms);
	sort!("a.sh_addr < b.sh_addr")(ssecs);
	sort!("a.st_value < b.st_value")(ssyms);
    }
    void parseFile(string filename)
    {
        parseStream(new BufferedFile(filename));
    }
    Elf32_Sym findSymbol(Elf32_Addr addr)
    {
	auto p = lowerBound!("a.st_value <= b")(ssyms, addr);
	failif(p == begin(ssyms) || addr >= p[-1].st_value + p[-1].st_size, "found no symbol containing the address");
	return p[-1];
    }
    Elf32_Shdr findSection(Elf32_Addr addr)
    {
	auto p = lowerBound!("a.sh_addr <= b")(ssecs, addr);
	failif(p == begin(ssecs) || addr >= p[-1].sh_addr + p[-1].sh_size, "found no section containing the address");
	return p[-1];
    }
    Elf32_Shdr section(Elf32_Sym sym) { return shdr(sym.st_shndx, "section of symbol " ~ name(sym)); }
    string name(Elf32_Sym sym) { return symstrtab.get(sym.st_name); }
    string name(Elf32_Shdr sec) { return secstrtab.get(sec.sh_name); }
}
